import logging
import re
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, List, NamedTuple, Optional, Union
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

from bs4 import BeautifulSoup, Tag
from cdp_backend.database.constants import (
    EventMinutesItemDecision,
    MatterStatusDecision,
    VoteDecision,
)
from cdp_backend.pipeline.ingestion_models import (
    Body,
    EventIngestionModel,
    EventMinutesItem,
    Matter,
    MinutesItem,
    Person,
    Session,
    SupportingFile,
    Vote,
)

from ..scraper_utils import (
    IngestionModelScraper,
    reduced_list,
    str_simplified,
    parse_static_file,
)

###############################################################################

log = logging.getLogger(__name__)

###############################################################################

SCRAPER_STATIC_DATA = parse_static_file(Path(__file__).parent / "portland-static.json")

###############################################################################

MATTER_ADOPTED_PATTERNS = [
    "accepted",
    "passed",
    "adopted",
    "confirmed",
]
MATTER_IN_PROG_PATTERNS = [
    "passed to",
    "placed on",
    "continued",
    "referred",
]

MINUTE_ITEM_PASSED_PATTERNS = [
    # NOTE: these words while have positive conotation,
    # does not mean the legistation was passed.
    # it indicates the item (or a report, etc.) was accepted to be discussed and voted.
    # "accepted",
    # "confirmed",
    # "adopted",
    "passed$",
]

###############################################################################


class WebPageSoup(NamedTuple):
    status: bool
    soup: Optional[BeautifulSoup] = None


def load_web_page(url: Union[str, Request]) -> WebPageSoup:
    """
    Load web page at url and return content soupified

    Parameters
    ----------
    url: str | urllib.request.Request
        Web page to load

    Returns
    -------
    result: WebPageSoup
        WebPageSoup.status = False if web page at url could not be loaded

    """
    try:
        with urlopen(url) as resp:
            return WebPageSoup(True, BeautifulSoup(resp.read(), "html.parser"))
    except URLError or HTTPError as e:
        log.error(f"Failed to open {url}: {str(e)}")

    return WebPageSoup(False)


def make_efile_url(efile_page_url: str) -> str:
    """
    Helper function to get file download link
    on a Portland EFile hosting web page

    Parameters
    ----------
    efile_page_url: str
        URL to Portland efile hosting web page
        e.g. https://efiles.portlandoregon.gov/record/14803529

    Returns
    -------
    efile url: str
        URL to the file itself
        e.g. https://efiles.portlandoregon.gov/record/14803529/File/Document
    """
    if not efile_page_url.endswith("/"):
        efile_page_url += "/"
    return f"{efile_page_url}File/Document"


def get_disposition(minute_section: Tag) -> str:
    """
    Return disposition string given within minute_section <div>
    on the event web page

    Parameters
    ----------
    minute_section: Tag
        <div> within event web page for a given event minute item

    Returns
    -------
    disposition: str
        Disposition string for the event minute item
        e.g. Accepted, Passed, Placed on file
    """
    result_status_element_sibling = minute_section.find(
        "div", text=re.compile("Disposition"), attrs={"class": "field__label"}
    )
    result_status_element = result_status_element_sibling.next_sibling
    return result_status_element.text


def disposition_to_minute_decision(
    disposition: str,
) -> Optional[EventMinutesItemDecision]:
    """
    Decide EventMinutesItemDecision constant from event minute item disposition

    Parameters
    ----------
    disposition: str
        Disposition event web page for a given item
        e.g. Passed, Continued

    Returns
    -------
    decision: Optional[EventMinutesItemDecision]

    See Also
    --------
    MINUTE_ITEM_PASSED_PATTERNS
    """
    for pattern in MINUTE_ITEM_PASSED_PATTERNS:
        if re.search(pattern, disposition, re.I):
            return EventMinutesItemDecision.PASSED
    return None


def separate_name_from_title(title_and_name: str) -> str:
    """
    Return just name

    Parameters
    ----------
    title_and_name: str
        e.g. Mayor Ted Wheeler

    Returns
    -------
    name: str
        tile_name_name with first word removed e.g. Ted Wheeler

    Notes
    -----
    first word in title_and_name is presumed to be title
    """
    # title_and_name:
    #   The title (Mayor of Commissioner) and name of a Portland City Commission
    #   member.
    #   e.g., Mayor Ted Wheeler, Commissioner Carmen Rubio

    name_index = title_and_name.find(" ")
    return title_and_name[name_index + 1 :]


class PortlandScraper(IngestionModelScraper):
    def __init__(self):
        super().__init__(timezone="America/Los_Angeles")

    def get_person(self, name: str) -> Person:
        """
        Return matching Person from portland-static.json

        Parameters
        ----------
        name: str
            Person full name

        Returns
        -------
        person: Person
            Matching Person from portland-static.json

        Raises
        ------
        KeyError
            If name does not exist in portland-static.json

        References
        ----------
        portland-static.json
        """
        if name not in SCRAPER_STATIC_DATA.persons:
            raise KeyError(f"{name} is unknown. Please update portland-static.json")

        return SCRAPER_STATIC_DATA.persons[name]

    def get_doc_number(self, minute_section: Tag, event_page: BeautifulSoup) -> str:
        """
        Find the document number in the minute_section.

        Parameters
        ----------
        minute_section: Tag
            <div> within event web page for a given event minute item
        event_page: BeautifulSoup
            The entire page where the event is found

        Returns
        -------
        doc_number: str
            The document number in the minute_section
            If this is null, return the section top number with the year
        """
        # Find document number
        doc_number_element_sibling = minute_section.find(
            "div", text=re.compile("Document number"), attrs={"class": "field__label"}
        )

        # If there is no document number, skip this minute item
        if doc_number_element_sibling is None:
            return self.get_section_top_number(minute_section, event_page)
        doc_number_element = doc_number_element_sibling.next_sibling
        doc_number = doc_number_element.find("div", class_="field__item").text.strip()
        return doc_number

    def get_section_top_number(
        self, minute_section: Tag, event_page: BeautifulSoup
    ) -> str:
        """
        Find the top section number in the minute_section.

        Parameters
        ----------
        minute_section: Tag
            <div> within event web page for a given event minute item
        event_page: BeautifulSoup
            The entire page where the event is found

        Returns
        -------
        doc_number: str
            The top section number in the minute_section, with the year appended at the
            end
        """
        agenda_name = event_page.find("title").text.strip()
        base_minute_section = minute_section.find("h4").text.strip()
        if agenda_name is not None:
            return (
                base_minute_section
                + "-"
                + agenda_name[agenda_name.index(",") + 2 : agenda_name.index(",") + 6]
            )
        return base_minute_section

    def get_matter(
        self, minute_section: Tag, event_page: BeautifulSoup
    ) -> Optional[Matter]:
        """
        Make Matter from information in minute_section

        Parameters
        ----------
        minute_section: Tag
            <div> within event web page for a given event minute item
        event_page: BeautifulSoup
            The entire page where the event is found

        Returns
        -------
        matter: Optional[Matter]
            Matter if required information could be parsed from minute_section
        """

        # Find title
        title_div = minute_section.find("div", class_="council-document__title")

        matter_type = None
        matter_title = None

        if title_div is not None:
            matter_title = title_div.find("a").text.strip()

            # Find type
            title_div.find("a").clear()
            matter_type = title_div.text.strip()
            if matter_type[0] == "(" and matter_type[-1] == ")":
                matter_type = matter_type[1:-1]
        else:
            matter_title = (
                minute_section.find("div", class_="field--name-field-disposition-notes")
                .children.__next__()
                .text.strip()
            )
            matter_type = matter_title[
                matter_title.rindex("(") + 1 : matter_title.rindex(")")
            ]
            matter_title = matter_title[0 : matter_title.rindex("(") - 1]

        # Find result status
        result_status = get_disposition(minute_section)
        # strings like "passed to second reading" is better to catch
        # before searching for "passed".
        # so test for IN_PROGRESS first.
        for pattern in MATTER_IN_PROG_PATTERNS:
            if re.search(pattern, result_status, re.I):
                result_status = MatterStatusDecision.IN_PROGRESS
                break
        else:
            for pattern in MATTER_ADOPTED_PATTERNS:
                if re.search(pattern, result_status, re.I):
                    result_status = MatterStatusDecision.ADOPTED
                    break
            else:
                result_status = None

        # Find the sponsors
        sponsor_element_uncle = minute_section.find(
            "div", text=re.compile("Introduced by"), attrs={"class": "field__label"}
        )
        sponsor_list = None
        if sponsor_element_uncle is not None:
            sponsor_element_parent = sponsor_element_uncle.next_sibling
            sponsor_elements = sponsor_element_parent.find_all(
                "div", class_="field__item"
            )
            sponsor_list = reduced_list(
                [
                    self.get_person(
                        separate_name_from_title(sponsor_element.text.strip())
                    )
                    for sponsor_element in sponsor_elements
                ]
            )

        return self.get_none_if_empty(
            Matter(
                matter_type=matter_type,
                name=self.get_doc_number(minute_section, event_page),
                sponsors=sponsor_list,
                title=matter_title,
                result_status=result_status,
            ),
        )

    def get_supporting_files(
        self, minute_section: Tag
    ) -> Optional[List[SupportingFile]]:
        """
        Return SupportingFiles for a given EventMinutesItem

        Parameters
        ----------
        minute_section: Tag
            <div> within event web page for a given event minute item

        Returns
        -------
        supporting files: Optional[List[SupportingFile]]

        Notes
        -----
        Follow hyperlink to go to minutes item details page.
        On the details page look for directly-linked files
        and externally-hosted efiles.

        See Also
        --------
        make_efile_url()
        """
        try:
            # on the event page, event minute item titles are listed
            # in <div> with a particular class attribute.
            # so look for the minute_item_index-th such <div> on the event page
            div = minute_section.find(
                "div", class_="field--label-hidden council-document__title"
            )
            # <a href="/council/documents/communication/placed-file/295-2021">
            details_url = f'https://www.portland.gov{div.find("a")["href"]}'
        except (AttributeError, TypeError):
            # minute_section.find() or div.find() failed
            return None

        # load the mintues item details page that may have links to supporting files
        details_soup = load_web_page(details_url)
        if not details_soup.status:
            return None

        supporting_files: List[SupportingFile] = []
        # first, try to get Documents and Exhibits and Impact Statement
        # these will contain links to files
        for div in details_soup.soup.find_all(
            "div",
            class_=re.compile(
                "field field--label-above field--name-field-"
                "((documents-and-exhibits)|(file-impact-statement)) field--type-file"
            ),
        ):
            supporting_files.extend(
                [
                    self.get_none_if_empty(
                        SupportingFile(
                            name=str_simplified(
                                re.sub(
                                    r"\s*download\s+file\s*",
                                    "",
                                    link.text,
                                    flags=re.IGNORECASE,
                                )
                            ),
                            uri=f'https://www.portland.gov{link["href"]}',
                        )
                    )
                    # <a href="/sites/...pdf"><span>Download file</span>
                    # <i class="fas fa-file-alt"></i>Exhibit A</a>
                    for link in div.find_all("a")
                ]
            )

        # finally parse for efile links
        # these are hosted yet on another web page; always start with https://efiles
        supporting_files.extend(
            [
                self.get_none_if_empty(
                    SupportingFile(
                        name=str_simplified(link.string),
                        uri=make_efile_url(link["href"]),
                    )
                )
                for link in details_soup.soup.find_all(
                    "a", href=re.compile(r"https:\/\/efiles.+")
                )
            ]
        )

        # remove any Nones
        return reduced_list(supporting_files)

    def get_votes(self, minute_section: Tag) -> Optional[List[Vote]]:
        """
        Look for 'Votes:' in minute_section and
        create a Vote object for each line

        Parameters
        ----------
        minute_section: Tag
            <div> within event web page for a given event minute item

        Returns
        -------
        votes: Optional[List[Vote]]
            Votes for corresponding event minute item if found
        """
        vote_element_uncle = minute_section.find(
            "div", text=re.compile("Votes"), attrs={"class": "field__label"}
        )
        if vote_element_uncle is None:
            return None
        vote_element_parent = vote_element_uncle.next_sibling
        vote_elements = vote_element_parent.find_all("div", class_="relation--type-")
        vote_list = []
        for vote_element in vote_elements:
            vote = vote_element.text.strip()
            # at this point vote string is like
            # Commissioner Jo Ann Hardesty Absent
            # Commissioner Mingus Mapps Yea
            is_absent = "absent" in vote.lower()
            vote = re.sub("absent", "", vote, flags=re.I)

            if "yea" in vote.lower():
                vote = re.sub("yea", "", vote, flags=re.I)
                decision = VoteDecision.APPROVE
                if is_absent:
                    decision = VoteDecision.ABSENT_APPROVE
            elif "nay" in vote.lower():
                vote = re.sub("nay", "", vote, flags=re.I)
                decision = VoteDecision.REJECT
                if is_absent:
                    decision = VoteDecision.ABSENT_REJECT
            elif is_absent:
                decision = VoteDecision.ABSENT_NON_VOTING
            else:
                decision = None

            # at this point any decision token like yea has been removed from vote
            name = separate_name_from_title(vote.strip())
            vote_list.append(
                self.get_none_if_empty(
                    Vote(decision=decision, person=self.get_person(name))
                )
            )

        return reduced_list(vote_list)

    def get_event_minutes(
        self, event_page: BeautifulSoup
    ) -> Optional[List[EventMinutesItem]]:
        """
        Make EventMinutesItem from each relation--type-agenda-item <div>
        on event_page

        Parameters
        ----------
        event_page: BeautifulSoup
            Web page for the meeting loaded as a bs4 object

        Returns
        -------
        event minute items: Optional[List[EventMinutesItem]]
        """
        minute_sections = event_page.find_all(
            "div", class_="relation--type-agenda-item"
        )
        event_minute_items = []
        for minute_section in minute_sections:
            matter = self.get_matter(minute_section, event_page)
            if matter is not None:
                minute_name = self.get_doc_number(minute_section, event_page)
                if minute_name is None:
                    minute_name = self.get_doc_number(minute_section, event_page)
                minutes_item = self.get_none_if_empty(
                    MinutesItem(name=minute_name, description=matter.title)
                )
            else:
                minutes_item = None

            event_minute_items.append(
                self.get_none_if_empty(
                    EventMinutesItem(
                        decision=disposition_to_minute_decision(
                            get_disposition(minute_section)
                        ),
                        matter=matter,
                        minutes_item=minutes_item,
                        supporting_files=self.get_supporting_files(minute_section),
                        votes=self.get_votes(minute_section),
                    )
                )
            )

        return reduced_list(event_minute_items)

    def get_sessions(self, event_page: BeautifulSoup) -> Optional[List[Session]]:
        """
        Parse meeting video URIs from event_page,
        return Session for each video found.

        Parameters
        ----------
        event_page: BeautifulSoup
            Web page for the meeting loaded as a bs4 object

        Returns
        -------
        sessions: Optional[List[Session]]
            Session for each video found on event_page
        """
        # each session's meta data is given in <div class="session-meta">
        # including youtube video url for the session, if available
        # <div class="session-meta">
        # ...
        # <time class="datetime">Wednesday, December 15, 2021 9:30 am</time>
        # ...
        # <iframe src="https://www.youtube.com/...">

        sessions: List[Session] = []
        session_index = 0

        for session_div in event_page.find_all("div", class_="session-meta"):
            session_time = session_div.find("time", class_="datetime")
            # plenty of sessions have no video listed so must check.
            # recall we require video_uri for a valid Session.
            video_iframe = session_div.find("iframe", src=re.compile(".*youtube.*"))

            if session_time and video_iframe:
                sessions.append(
                    self.get_none_if_empty(
                        Session(
                            session_datetime=self.localize_datetime(
                                datetime.strptime(
                                    session_time.string,
                                    "%A, %B %d, %Y %I:%M %p",
                                )
                            ),
                            session_index=session_index,
                            video_uri=video_iframe["src"].split("?")[0],
                        )
                    )
                )

            session_index += 1

        return reduced_list(sessions)

    def get_agenda_uri(self, event_page: BeautifulSoup) -> Optional[str]:
        """
        Find the uri for the file containing the agenda for a Portland, OR city
        council meeting

        Parameters
        ----------
        event_page: BeautifulSoup
            Web page for the meeting loaded as a bs4 object

        Returns
        -------
        agenda_uri: Optional[str]
            The uri for the file containing the meeting's agenda
        """
        agenda_uri_element = event_page.find(
            "a", text=re.compile("Disposition Agenda"), attrs={"class": "btn-cta"}
        )
        if agenda_uri_element is not None:
            return make_efile_url(agenda_uri_element["href"])
        parent_agenda_uri_element = event_page.find("div", {"class": "inline-flex"})
        if parent_agenda_uri_element is not None:
            agenda_uri_element = parent_agenda_uri_element.find("a")
        else:
            return None
        if agenda_uri_element is not None:
            return f"https://www.portland.gov{agenda_uri_element['href']}"
        return None

    def get_event(self, event_time: datetime) -> Optional[EventIngestionModel]:
        """
        Portland, OR city council meeting information for a specific date

        Parameters
        ----------
        event_time: datetime
            Meeting date

        Returns
        -------
        Optional[EventIngestionModel]
            None if there was no meeting on event_time
            or information for the meeting did not meet minimal CDP requirements.
        """
        # try to load https://www.portland.gov/council/agenda/yyyy/m/d
        event_page = load_web_page(
            "https://www.portland.gov/council/agenda/"
            # we actually DON'T want to use strftime() because we must not zero-pad
            # e.g. for 2022/01/05, we MUST use 2022/1/5
            f"{event_time.year}/{event_time.month}/{event_time.day}"
        )
        if not event_page.status:
            # no meeting on requested day
            return None

        return self.get_none_if_empty(
            EventIngestionModel(
                agenda_uri=self.get_agenda_uri(event_page.soup),
                # NOTE: have not seen any specific body/bureau named on any agenda page
                body=Body(name="City Council"),
                event_minutes_items=self.get_event_minutes(event_page.soup),
                minutes_uri=None,
                sessions=self.get_sessions(event_page.soup),
            ),
        )

    def get_events(
        self,
        begin: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> List[EventIngestionModel]:
        """
        Portland, OR city council meeting information over given time span
        as List[EventIngestionModel]

        Parameters
        ----------
        begin: datetime, optional
            The timespan beginning datetime to query for events after.
            Default is 2 days from UTC now
        end: datetime, optional
            The timespan end datetime to query for events before.
            Default is UTC now

        Returns
        -------
        events: List[EventIngestionModel]

        References
        ----------
        https://www.portland.gov/council/agenda/all
        """
        if begin is None:
            begin = datetime.utcnow() - timedelta(days=2)
        if end is None:
            end = datetime.utcnow()

        return reduced_list(
            [
                self.get_event(begin + timedelta(days=day))
                for day in range((end - begin).days)
            ],
            # prefer to return [] over None to backend pipeline
            # for easier iterate there
            collapse=False,
        )


def get_portland_events(
    from_dt: Optional[datetime] = None,
    to_dt: Optional[datetime] = None,
    **kwargs: Any,
) -> List[EventIngestionModel]:
    """
    Public API for use in instances.__init__ so that this func can be attached
    as an attribute to cdp_scrapers.instances module.
    Thus the outside world like cdp-backend can get at this by asking for
    "get_portland_events".

    Parameters
    ----------
    from_dt: datetime, optional
        The timespan beginning datetime to query for events after.
        Default is 2 days from UTC now
    to_dt: datetime, optional
        The timespan end datetime to query for events before.
        Default is UTC now

    Returns
    -------
    events: List[EventIngestionModel]

    See Also
    --------
    cdp_scrapers.instances.__init__.py
    """
    scraper = PortlandScraper()
    return scraper.get_events(begin=from_dt, end=to_dt, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import re
import json
from typing import Dict
from urllib.request import urlopen
from pathlib import Path

from bs4 import BeautifulSoup

from cdp_backend.pipeline.ingestion_models import Person, Seat
from cdp_backend.database.constants import RoleTitle
from ..legistar_utils import (
    LegistarScraper,
)
from ..scraper_utils import str_simplified, parse_static_file

###############################################################################

log = logging.getLogger(__name__)

###############################################################################

STATIC_FILE_KEY_PERSONS = "persons"
STATIC_FILE_DEFAULT_PATH = Path(__file__).parent / "kingcounty-static.json"

###############################################################################


class KingCountyScraper(LegistarScraper):
    PYTHON_MUNICIPALITY_SLUG: str = "king_county"

    def __init__(self):
        """
        King County specific implementation of LegistarScraper.
        """
        super().__init__(
            client="kingcounty",
            timezone="America/Los_Angeles",
            ignore_minutes_item_patterns=[
                "This meeting also constitutes a meeting of the City Council",
                "In-person attendance is currently prohibited",
                "Times listed are estimated",
                "has been cancelled",
                "Deputy City Clerk",
                "Paste the following link into the address bar of your web browser",
                "HOW TO WATCH",
                "page break",
                "PUBLIC NOTICE",
                "There will be one public hearing on",
                "Consent Items",
                "SUBJECT TO A MOTION TO SUSPEND THE RULES TO TAKE ACTION WITHOUT REFERRAL TO COMMITTEE PURSUANT",  # noqa: E501
                "If you do not have access to the ZOOM application",
                "Executive Session: For 15 minutes, with action to follow",
                "on the agenda for procedural matters",
                "This is a mandatory referral to the",
                "Watch King County TV Channel 22",
            ],
            static_data=parse_static_file(STATIC_FILE_DEFAULT_PATH),
            role_replacements={
                "Boardmember": RoleTitle.MEMBER,
                "Mr.": RoleTitle.MEMBER,
                "Councilmemeber DELETE": RoleTitle.COUNCILMEMBER,
                "Vice-Chair": RoleTitle.VICE_CHAIR,
                "Council Member": RoleTitle.COUNCILMEMBER,
                "Policy Chair": RoleTitle.CHAIR,
            },
        )

    @staticmethod
    def get_static_person_info() -> Dict[str, Person]:
        """
        Scrape current council members information from kingcounty.gov

        Returns
        -------
        persons: Dict[str, Person]
            keyed by name

        Notes
        -----
        Parse https://kingcounty.gov/council/councilmembers/find_district.aspx
        that contains list of current council members name, position, contact info
        """
        # this page lists current council members
        with urlopen(
            "https://kingcounty.gov/council/councilmembers/find_district.aspx"
        ) as resp:
            soup = BeautifulSoup(resp.read(), "html.parser")

        # keyed by name
        persons: Dict[str, Person] = {}

        # there is a series of council member portrait pictures
        # marked by text "official portrait"

        # <a href="/council/dembowski.aspx"><strong>Rod Dembowski</strong><br/>
        # </a>District 1<br/>
        # 206-477-1001<br/>
        # <a href="mailto:...">rod.dembowski@kingcounty.gov </a><br/>
        # Member since: 2013<br/>
        # Current t<span style="color: rgb(0, 0, 0);">erm: 2018-2021</span><br/>
        # <a href="/~/media/...">Official portrait</a>
        # ...
        # repeat similar blob for the next council person
        # ...
        for picture in soup.find_all(
            "a", text=re.compile(".*official.*portrait.*", re.IGNORECASE)
        ):
            picture_uri = f"https://kingcounty.gov{str_simplified(picture['href'])}"

            # preceding 2 <a> tags have email and website url, in that order
            email_tag = picture.find_previous_sibling("a")
            email = str_simplified(email_tag.text)
            website_tag = email_tag.find_previous_sibling("a")
            website = f"https://kingcounty.gov{str_simplified(website_tag['href'])}"
            name = str_simplified(website_tag.text)

            # this area in plain text contains role and phone for this person
            # Rod Dembowski
            # District 1
            # 206-477-1001
            # rod.dembowski@kingcounty.gov
            parent_tag = picture.find_parent()
            # position number is the trailing number in the line after name
            # and the line after position is the telephone
            match = re.search(
                f".*{name}.*\\s+.*(?P<position>\\d+)\\s+(?P<phone>[\\d\\-\\(\\)]+)",
                parent_tag.text,
                re.IGNORECASE,
            )
            phone = match.group("phone")
            seat = Seat(name=f"Position {match.group('position')}")

            persons[name] = Person(
                name=name,
                picture_uri=picture_uri,
                email=email,
                website=website,
                phone=phone,
                seat=seat,
            )

        return persons

    @staticmethod
    def dump_static_info(file_path: Path) -> None:
        """
        Call this to save current council members information as Persons
        in json format to file_path.
        Intended to be called once every N years when the council changes.

        Parameters
        ----------
        file_path: Path
            output json file path
        """
        static_info_json = {STATIC_FILE_KEY_PERSONS: {}}
        for [name, person] in KingCountyScraper.get_static_person_info().items():
            # to allow for easy future addition of info other than Persons
            # save under top-level key "persons" in the file
            static_info_json[STATIC_FILE_KEY_PERSONS][name] = person.to_dict()

        with open(file_path, "wt") as dump:
            dump.write(json.dumps(static_info_json, indent=4))



from typing import Optional, Tuple
import selenium
import re
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from webdriver_manager.chrome import ChromeDriverManager
from cdp_backend.pipeline import ingestion_models
from cdp_backend.database import constants as db_constants
from datetime import datetime
import logging

log = logging.getLogger(__name__)
chrome_options = Options()
chrome_options.add_argument("--headless")


def get_single_person(
    driver: ChromeDriverManager, member_name: str
) -> ingestion_models.Person:
    """
    Get all the information fot one person
    Includes: role, seat, picture, phone, email

    Parameter:
    ----------------
    driver:
        webdriver calling the people's dictionary page
    member_name:
        person's name

    Returns:
    --------------
    ingestion_models
        the ingestion model for the person's part
    """
    log.info("start get active person ingestion model")
    seat_role = driver.find_element(By.CLASS_NAME, "titlewidget-subtitle").text
    member_role = db_constants.RoleTitle.MEMBER
    member_seat_area = None
    member_seat_name = db_constants.RoleTitle.COUNCILPRESIDENT
    if "President" in seat_role:
        member_role = db_constants.RoleTitle.COUNCILPRESIDENT
        member_seat_area = "Citywide"
    elif "Post" in seat_role:
        name_list = seat_role.split(" ")
        member_seat_name = "Post " + name_list[1]
        member_seat_area = "Citywide"
    else:
        member_seat_name = seat_role
    member_pic = driver.find_element(
        By.CSS_SELECTOR, ".image_widget img"
    ).get_attribute("src")
    temp_email = (
        driver.find_element(By.XPATH, "// a[contains(text(),'Click Here')]")
        .get_attribute("href")
        .split(":")
    )
    member_email = temp_email[1]
    try:
        member_details = driver.find_element(
            By.XPATH, "//*[contains(@id, 'widget_340_')]"
        ).text
    except (selenium.common.exceptions.NoSuchElementException):
        member_details = driver.find_element(
            By.XPATH, "//*[contains(@id, 'widget_437_')]"
        ).text
    detail_str = member_details.split("\n")
    phone_list = [s for s in detail_str if "P" in s]
    member_phone = phone_list[0].split(": ")[1]

    return ingestion_models.Person(
        name=member_name,
        is_active=True,
        email=member_email,
        phone=member_phone,
        picture_uri=member_pic,
        seat=ingestion_models.Seat(
            name=member_seat_name,
            electoral_area=member_seat_area,
            roles=[ingestion_models.Role(title=member_role)],
        ),
    )


def get_person() -> dict:
    """
    Put the informtion get by get_single_person() to dictionary

    Returns:
    --------------
    dictionary
        key: person's name
        value: person's ingestion model
    """
    log.info("start get all the person ingestion model")
    driver = webdriver.Chrome(
        options=chrome_options, service=Service(ChromeDriverManager().install())
    )
    driver.get("https://citycouncil.atlantaga.gov/council-members")
    members = driver.find_elements(By.XPATH, '//*[@id="leftNav_2_0_12"]/ul/li')
    person_dict = {}
    for member in members:
        link = member.find_element(By.TAG_NAME, "a").get_attribute("href")
        driver = webdriver.Chrome(
            options=chrome_options, service=Service(ChromeDriverManager().install())
        )
        driver.get(link)
        member_name = driver.find_element(By.CLASS_NAME, "titlewidget-title").text
        if "President" in member_name:
            member_name = member_name.split("President ")[1]
        else:
            current_name = re.match(
                r"([a-zA-Z]+)((\ {0,1}[a-zA-Z]+\.{0,1}\ )|(\ ))([a-zA-Z]+)", member_name
            )
            if current_name is not None:
                member_name = f"{current_name.group(1)} {current_name.group(5)}"
            else:
                raise ValueError("Person name could not be constructed.")
        member_model = get_single_person(driver, member_name)
        driver.quit()
        person_dict[member_name] = member_model
    driver.quit()
    return person_dict


def get_new_person(name: str) -> ingestion_models.Person:
    """
    Creates the person ingestion model for the people that are not recored

    Parameter:
    ----------------
    name:str
        the name of the person

    Returns:
    --------------
    ingestion model
        the person ingestion model for the newly appeared person
    """
    log.info("start get inactive person ingestion model")
    return ingestion_models.Person(name=name, is_active=False)


def convert_status_constant(decision: str) -> str:
    """
    Converts the matter result status to the exsiting constants

    Parameter:
    ----------------
    decision: str
        decision of the matter

    Returns:
    --------------
    db_constants
        result status constants
    """
    log.info("start convert result status for vote ingestion model")
    d_constant = decision
    if ("FAVORABLE" in decision) or ("ADOPTED" in decision) or ("ACCEPTED" in decision):
        d_constant = db_constants.MatterStatusDecision.ADOPTED
    elif (
        ("REFERRED" in decision)
        or ("RETURNED" in decision)
        or ("FILED" in decision)
        or ("Refer")
        or ("/" in decision)
    ):
        d_constant = db_constants.MatterStatusDecision.IN_PROGRESS
    else:
        raise ValueError("New Type")
    return d_constant


def assign_constant(
    driver: ChromeDriverManager,
    i: int,
    j: int,
    vote_decision: str,
    voting_list: list,
    body_name: str,
    PERSONS: dict,
):
    """
    Assign constants and add Vote to the ingestion models based on the vote decision

    Parameter:
    ----------------
    driver:webdriver
        webdriver of the matter page
    i: int
        tr[i] is the matter we are looking at
    j: int
        the row number of the information in a matter that we are looking at
    vote_decision: str
        the vote decision constant of the vote decision
    voting_list: list
        the list that contains vote ingestion models
    body_name: str
        the body name of the current meeting
    PERSONS: dict
        Dict[str, ingestion_models.Person]
    """
    log.info("start get vote ingestion model for one type of decision")
    v_res = driver.find_element(
        By.XPATH,
        '//*[@id="ContentPlaceHolder1_divHistory"]/div/table/tbody/tr['
        + str(i + 1)
        + "]/td/table/tbody/tr["
        + str(j)
        + "]/td[2]",
    ).text
    res_list = v_res.split(", ")
    n: str = ""
    for p in res_list:
        if "President" in p:
            n = p.split("President ")[1]
        else:
            n_temp = re.match(
                r"([a-zA-Z]+)((\ {0,1}[a-zA-Z]+\.{0,1}\ )|(\ ))([a-zA-Z]+)", p
            )
            if n_temp is not None:
                n = f"{n_temp.group(1)} {n_temp.group(5)}"
            else:
                raise ValueError("Person name could not be constructed.")
        person = get_new_person(n)
        if n in PERSONS:
            person = PERSONS[n]
            if person.seat is not None:
                if person.seat.roles is not None:
                    person.seat.roles[0].body = ingestion_models.Body(
                        body_name, is_active=True
                    )
                    if body_name == "City Council":
                        if (
                            person.seat.roles[0].title
                            != db_constants.RoleTitle.COUNCILPRESIDENT
                        ):
                            person.seat.roles[
                                0
                            ].title = db_constants.RoleTitle.COUNCILMEMBER
        voting_list.append(
            ingestion_models.Vote(
                person=person,
                decision=vote_decision,
            )
        )
    return voting_list


def get_voting_result(
    driver: ChromeDriverManager,
    sub_sections_len: int,
    i: int,
    body_name: str,
    PERSONS: dict,
) -> list:
    """
    Scrapes and converts the voting decisions to the exsiting constants

    Parameter:
    ----------------
    driver:webdriver
        webdriver of the matter page
    sub_sections_len: int
        the row number in the block under the matter for the current date
    i: int
        tr[i] is the matter we are looking at
    body_name: str
        the body name of the current meeting
    PERSONS: dict
        Dict[str, ingestion_models.Person]

    Returns:
    --------------
    list
        contains the Vote ingestion model for each person
    """
    log.info("start get the vote ingestion model for a matter")
    voting_list: list[ingestion_models.Vote] = []
    for j in range(1, sub_sections_len + 1):
        sub_content = driver.find_element(
            By.XPATH,
            '//*[@id="ContentPlaceHolder1_divHistory"]/div/table/tbody/tr['
            + str(i + 1)
            + "]/td/table/tbody/tr["
            + str(j)
            + "]",
        )
        sub_content_role = sub_content.find_element(By.CLASS_NAME, "Role").text
        if "AYES" in sub_content_role:
            vote_decision = db_constants.VoteDecision.APPROVE
            assign_constant(
                driver, i, j, vote_decision, voting_list, body_name, PERSONS
            )
        if "NAYS" in sub_content_role:
            vote_decision = db_constants.VoteDecision.REJECT
            assign_constant(
                driver, i, j, vote_decision, voting_list, body_name, PERSONS
            )
        if (
            "ABSENT" in sub_content_role
            or "AWAY" in sub_content_role
            or "EXCUSED" in sub_content_role
        ):
            vote_decision = db_constants.VoteDecision.ABSENT_NON_VOTING
            assign_constant(
                driver, i, j, vote_decision, voting_list, body_name, PERSONS
            )
        if "ABSTAIN" in sub_content_role:
            vote_decision = db_constants.VoteDecision.ABSTAIN_NON_VOTING
            assign_constant(
                driver, i, j, vote_decision, voting_list, body_name, PERSONS
            )
    return voting_list


def get_matter_status(driver: ChromeDriverManager, i: int) -> Tuple[list, str]:
    """
    Find the matter result status

    Parameter:
    ----------------
    driver:webdriver
        webdriver of the matter page
    i: int
        tracker used to loop the rows in the matter page

    Returns:
    --------------
    sub_sections: element
        the block under the matter for the current date
    decision_constant: element
        the matter decision constant
    """
    log.info("start get reslut status for a matter")
    result = driver.find_element(
        By.XPATH,
        '//*[@id="ContentPlaceHolder1_divHistory"]/div/table/tbody/tr['
        + str(i + 1)
        + "]/td/table",
    )
    decision = result.find_element(By.CLASS_NAME, "Result").text
    sub_sections = result.find_elements(
        By.XPATH,
        '//*[@id="ContentPlaceHolder1_divHistory"]/div/table/tbody/tr['
        + str(i + 1)
        + "]/td/table/tbody/tr",
    )
    status_constant = convert_status_constant(decision)
    return sub_sections, status_constant


def parse_single_matter(
    driver: ChromeDriverManager,
    test: str,
    item: str,
    body_name: str,
    s_word_formated: datetime,
    PERSONS: dict,
) -> ingestion_models.EventMinutesItem:
    """
    Get the minute items that contains a matter

    Parameter:
    ----------------
    driver:webdriver
        webdriver of the matter page
    matter:element
        the matter we are looking at
    body_name: str
        the body name of the current meeting
    s_word_formated: datetime
        the date of the current meeting
    PERSONS: dict
        Dict[str, ingestion_models.Person]

    Returns:
    --------------
    ingestion model
        minutes ingestion model with the matters information
    """
    log.info("start get ingestion model for a matter")
    voting_list = []
    matter_name = item[0:9]  # name of the matter eg. "22-C-5024", "22-R-3404"
    matter_title = item[
        12:
    ]  # the paragraph the describes the matter eg. "A COMMUNICATION FROM ..."
    matter_type = (
        " "  # the type of the matter eg. "COMMUNICATION", "SUBSTITUTE ORDINANCE"
    )
    matter_type_temp = re.split(" BY| FROM", matter_title)[0]
    matter_type_list = re.split("A |AN ", matter_type_temp)
    if len(matter_type_list) > 1:
        matter_type = matter_type_list[1]
    link = driver.find_element("link text", item)
    link.click()
    # get to the specific page for each matter
    s_matter = WebDriverWait(driver, 10).until(
        EC.presence_of_all_elements_located(
            (
                By.XPATH,
                '//*[@id="ContentPlaceHolder1_divHistory"]/div/table/tbody/tr',
            )
        )
    )
    sponsor_raw = driver.find_element(
        By.XPATH, '//*[@id="tblLegiFileInfo"]/tbody/tr[1]/td[2]'
    ).text
    sponsor_list = sponsor_raw.split(", ")
    sponsors: Optional[list[ingestion_models.Person]] = []
    if sponsors is not None:
        for s in sponsor_list:
            if "District" in s:
                current_temp = s.split(" ")[2:]
                current_temp2 = " ".join(current_temp)
                current_name = re.match(
                    r"([a-zA-Z]+)((\ {0,1}[a-zA-Z]+\.{0,1}\ )|(\ ))([a-zA-Z]+)",
                    current_temp2,
                )
                if current_name is not None:
                    current = f"{current_name.group(1)} {current_name.group(5)}"
                else:
                    raise ValueError("Person name could not be constructed.")
                if current in PERSONS:
                    sponsors.append(PERSONS[current])
                else:
                    sponsors.append(get_new_person(current))
            elif "Post" in s:
                current_temp = s.split("Large ")[1]
                current_name = re.match(
                    r"([a-zA-Z]+)((\ {0,1}[a-zA-Z]+\.{0,1}\ )|(\ ))([a-zA-Z]+)",
                    current_temp,
                )
                if current_name is not None:
                    current = f"{current_name.group(1)} {current_name.group(5)}"
                else:
                    raise ValueError("Person name could not be constructed.")
                if current in PERSONS:
                    sponsors.append(PERSONS[current])
                else:
                    sponsors.append(get_new_person(current))
            elif "President" in s:
                current_temp = s.split("President ")[1]
                current_name = re.match(
                    r"([a-zA-Z]+)((\ {0,1}[a-zA-Z]+\.{0,1}\ )|(\ ))([a-zA-Z]+)",
                    current_temp,
                )
                if current_name is not None:
                    current = f"{current_name.group(1)} {current_name.group(5)}"
                else:
                    raise ValueError("Person name could not be constructed.")
                if current in PERSONS:
                    sponsors.append(PERSONS[current])
                else:
                    sponsors.append(get_new_person(current))
        s_rows = len(s_matter)
        for i in range(1, s_rows + 1, 2):
            header = driver.find_element(
                By.XPATH,
                '//*[@id="ContentPlaceHolder1_divHistory"]/div/table/tbody/tr['
                + str(i)
                + "]",
            )
            date = header.find_element(By.CLASS_NAME, "Date").text
            date_formated = datetime.strptime(date[:-6], "%b %d, %Y %I:%M %p")
            if s_word_formated == date_formated:  # match the current meeting date
                sub_sections, status = get_matter_status(
                    driver, i
                )  # get the decision and result_status of the matter
                if status in (
                    db_constants.MatterStatusDecision.IN_PROGRESS,
                    db_constants.MatterStatusDecision.ADOPTED,
                ):
                    decision = db_constants.EventMinutesItemDecision.PASSED
                else:
                    decision = db_constants.EventMinutesItemDecision.FAILED
                if "[" in test:
                    voting_list = get_voting_result(
                        driver, len(sub_sections), i, body_name, PERSONS
                    )
        if len(sponsors) != 0:
            return ingestion_models.EventMinutesItem(
                minutes_item=ingestion_models.MinutesItem(
                    name=matter_name.title(), description=matter_title
                ),
                matter=ingestion_models.Matter(
                    matter_name,
                    matter_type=matter_type,
                    title=matter_title,
                    result_status=status,
                    sponsors=sponsors,
                ),
                decision=decision,
                votes=voting_list,
            )
    return ingestion_models.EventMinutesItem(
        minutes_item=ingestion_models.MinutesItem(
            name=matter_name.title(), description=matter_title
        ),
        matter=ingestion_models.Matter(
            matter_name,
            matter_type=matter_type,
            title=matter_title,
            result_status=status,
        ),
        decision=decision,
        votes=voting_list,
    )


def parse_event(url: str) -> ingestion_models.EventIngestionModel:
    """
    Scrapes all the information for a meeting

    Parameter:
    ----------------
    url:str
        the url of the meeting that we want to scrape

    Returns:
    --------------
    ingestion model
        the ingestion model for the meeting
    """
    log.info("start get ingestion model for a event")

    MINUTE_INDEX = [chr(i) for i in range(ord("A"), ord("Z") + 1)]
    PERSONS = get_person()

    driver = webdriver.Chrome(
        options=chrome_options, service=Service(ChromeDriverManager().install())
    )
    driver.get(url)

    WebDriverWait(driver, 10).until(
        EC.presence_of_all_elements_located(
            (By.XPATH, '//*[@id="MeetingDetail"]/tbody/tr')
        )
    )

    body_name = driver.find_element(
        By.ID, "ContentPlaceHolder1_lblMeetingGroup"
    ).text  # body name
    if body_name == "Atlanta City Council":
        body_name = "City Council"
    s_word = driver.find_element(By.ID, "ContentPlaceHolder1_lblMeetingDate").text
    s_word_formated = datetime.strptime(s_word, "%m/%d/%Y %I:%M %p")
    video_link = driver.find_element(By.ID, "MediaPlayer1_html5_api").get_attribute(
        "src"
    )  # video link (mp4)

    event_minutes_items = []
    i = 1

    while (
        len(
            driver.find_elements(
                By.XPATH, '//*[@id="MeetingDetail"]/tbody/tr[' + str(i) + "]"
            )
        )
        != 0
    ):
        try:
            if (
                len(
                    driver.find_elements(
                        By.XPATH,
                        '//*[@id="MeetingDetail"]/tbody/tr['
                        + str(i)
                        + "]/td[1]/strong",
                    )
                )
            ) != 0 and (
                len(
                    driver.find_element(
                        By.XPATH,
                        '//*[@id="MeetingDetail"]/tbody/tr['
                        + str(i)
                        + "]/td[1]/strong",
                    ).text
                )
            ) != 0:
                if (
                    driver.find_element(
                        By.XPATH,
                        '//*[@id="MeetingDetail"]/tbody/tr['
                        + str(i)
                        + "]/td[1]/strong",
                    ).text
                )[0] in MINUTE_INDEX:
                    if (
                        len(
                            driver.find_elements(
                                By.XPATH,
                                '//*[@id="MeetingDetail"]/tbody/tr['
                                + str(i + 1)
                                + "]/td[3]/span",
                            )
                        )
                        == 0
                    ):
                        minute_title = driver.find_element(
                            By.XPATH,
                            '//*[@id="MeetingDetail"]/tbody/tr[' + str(i) + "]/td[2]",
                        ).text
                        minute_model = ingestion_models.EventMinutesItem(
                            minutes_item=ingestion_models.MinutesItem(
                                minute_title.title()
                            )
                        )
                        event_minutes_items.append(minute_model)
            elif (
                len(
                    driver.find_elements(
                        By.XPATH,
                        '//*[@id="MeetingDetail"]/tbody/tr[' + str(i) + "]/td[3]/span",
                    )
                )
            ) != 0:
                matter = driver.find_element(
                    By.XPATH, '//*[@id="MeetingDetail"]/tbody/tr[' + str(i) + "]/td[3]"
                )
                test = matter.find_element(By.CLASS_NAME, "ItemVoteResult").text
                item = matter.find_element(By.CLASS_NAME, "AgendaOutlineLink").text
                if len(item) != 0:
                    matter_model = parse_single_matter(
                        driver, test, item, body_name, s_word_formated, PERSONS
                    )
                    event_minutes_items.append(matter_model)
            elif (
                len(
                    driver.find_elements(
                        By.XPATH,
                        '//*[@id="MeetingDetail"]/tbody/tr[' + str(i) + "]/td[6]/span",
                    )
                )
            ) != 0:
                matter = driver.find_element(
                    By.XPATH, '//*[@id="MeetingDetail"]/tbody/tr[' + str(i) + "]/td[6]"
                )
                test = matter.find_element(By.CLASS_NAME, "ItemVoteResult").text
                item = matter.find_element(By.CLASS_NAME, "AgendaOutlineLink").text
                if len(item) != 0:
                    matter_model = parse_single_matter(
                        driver, test, item, body_name, s_word_formated, PERSONS
                    )
                    event_minutes_items.append(matter_model)
            i += 1
        except (
            selenium.common.exceptions.NoSuchElementException,
            selenium.common.exceptions.TimeoutException,
        ):
            i += 1
            continue

    agenda_link = driver.find_element(
        By.ID, "ContentPlaceHolder1_hlPublicAgendaFile"
    ).get_attribute("oldhref")
    minutes_link = driver.find_element(
        By.ID, "ContentPlaceHolder1_hlPublicMinutesFile"
    ).get_attribute("oldhref")

    driver.quit()

    return ingestion_models.EventIngestionModel(
        body=ingestion_models.Body(body_name, is_active=True),
        sessions=[
            ingestion_models.Session(
                video_uri=video_link,
                session_index=0,
                session_datetime=s_word_formated,
            )
        ],
        event_minutes_items=event_minutes_items,
        agenda_uri="https://atlantacityga.iqm2.com/Citizens/" + agenda_link,
        minutes_uri="https://atlantacityga.iqm2.com/Citizens/" + minutes_link,
    )


def get_year(driver: ChromeDriverManager, url: str, from_dt: datetime) -> str:
    """
    Navigate to the year that we are looking for

    Parameter:
    ----------------
    driver:webdriver
        empty webdriver
    url:str
        the url of the calender page

    Returns:
    --------------
    link:str
        the link to the calender of the year that we are looking for
    """
    log.info("start get the current year's calender page")
    driver.get(url)
    dates = driver.find_element(By.ID, "ContentPlaceHolder1_lblCalendarRange")
    link_temp = dates.find_element(
        By.XPATH, ("//*[text()='" + str(from_dt.year) + "']")
    ).get_attribute("href")
    link = "https://atlantacityga.iqm2.com" + link_temp
    return link


def get_date(
    driver: ChromeDriverManager, url: str, from_dt: datetime, to_dt: datetime
) -> list:
    """
    Get a list of ingestion models for the meetings hold during the selected time range

    Parameter:
    ----------------
    driver:webdriver
        empty webdriver
    url:str
        the url of the calender page
    from_dt:
        the begin date
    to_date:
        the end date

    Returns:
    --------------
    list
        all the ingestion models for the selected date range
    """
    log.info("start calling parse_event for a signle meeting")
    driver.get(url)
    dates = driver.find_elements(By.CLASS_NAME, "RowTop")
    events = []
    for current_date in dates:
        current_meeting_date = current_date.find_element(By.CLASS_NAME, "RowLink")
        current_meeting_time = datetime.strptime(
            current_meeting_date.text, "%b %d, %Y %I:%M %p"
        )
        if from_dt <= current_meeting_time <= to_dt:
            link_temp = current_date.find_element(
                By.CSS_SELECTOR, ".WithoutSeparator a"
            ).get_attribute("onclick")
            link = "https://atlantacityga.iqm2.com" + link_temp[23:-3]
            event = parse_event(link)
            events.append(event)
        else:
            continue
    driver.quit()
    return events


def get_events(from_dt: datetime, to_dt: datetime) -> list:
    """
    gets the right calender link
    feed it to the function that get a list of ingestion models

    Parameter:
    ----------------
    from_dt:
        the begin date
    to_date:
        the end date

    Returns:
    --------------
    list
        all the ingestion models for the selected date range
    """
    log.info("start a date range and run all the functions")
    driver = webdriver.Chrome(
        options=chrome_options, service=Service(ChromeDriverManager().install())
    )
    web_url = "https://atlantacityga.iqm2.com/Citizens/Calendar.aspx?Frame=Yes"
    driver.get(web_url)
    if from_dt.year != datetime.today().year:
        web_url = get_year(driver, web_url, from_dt)
    events = get_date(driver, web_url, from_dt, to_dt)
    return events



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime
from typing import List

from cdp_backend.pipeline.ingestion_models import EventIngestionModel

###############################################################################


def get_events(
    from_dt: datetime,
    to_dt: datetime,
    **kwargs,
) -> List[EventIngestionModel]:
    """
    Get all events for the provided timespan.

    Parameters
    ----------
    from_dt: datetime
        Datetime to start event gather from.
    to_dt: datetime
        Datetime to end event gather at.

    Returns
    -------
    events: List[EventIngestionModel]
        All events gathered that occured in the provided time range.

    Notes
    -----
    As the implimenter of the get_events function, you can choose to ignore the from_dt
    and to_dt parameters. However, they are useful for manually kicking off pipelines
    from GitHub Actions UI.
    """

    # Your implementation here
    return []



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Callable, Dict, List, NamedTuple, Optional
from bs4 import BeautifulSoup
from cdp_backend.pipeline.ingestion_models import (
    Body,
    Person,
    Seat,
)

###############################################################################


class ContentURIs(NamedTuple):
    video_uri: Optional[str]
    caption_uri: Optional[str] = None


class ScraperStaticData(NamedTuple):
    seats: Dict[str, Seat] = None
    primary_bodies: Dict[str, Body] = None
    persons: Dict[str, Person] = None


LegistarContentParser = Callable[[str, BeautifulSoup], Optional[List[ContentURIs]]]
"""
Function that returns URLs for videos and captions
from a Legistar/Granicus-hosted video web page

Parameters
----------
client: str
    Which legistar client to target. Ex: "seattle"
video web page: BeautifulSoup
    Video web page loaded into bs4

Returns
-------
uris: Optional[List[ContentURIs]]
    URIs for video and optional caption

See Also
--------
cdp_scrapers.legistar_content_parsers
cdp_scrapers.legistar_utils.get_legistar_content_uris()
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import re
import json
import requests
import urllib3
import warnings
from typing import Dict, List, Optional
from urllib.error import HTTPError, URLError
from urllib.request import urlopen
from urllib.parse import urlsplit, parse_qs, quote_plus
from pathlib import Path
from datetime import datetime

from bs4 import BeautifulSoup

from cdp_backend.pipeline.ingestion_models import Person, Seat

from ..legistar_utils import (
    LEGISTAR_EV_SITE_URL,
    LEGISTAR_SESSION_DATE,
    LegistarScraper,
)
from ..scraper_utils import str_simplified, parse_static_file
from ..types import ContentURIs

###############################################################################

log = logging.getLogger(__name__)

###############################################################################

STATIC_FILE_KEY_PERSONS = "persons"
STATIC_FILE_DEFAULT_PATH = Path(__file__).parent / "seattle-static.json"

# we have discovered the city clerk accidentally entered Daniel Strauss
# instead of the correct Dan Strauss for a few events
PERSON_ALIASES = {"Dan Strauss": set(["Daniel Strauss"])}

###############################################################################


class SeattleScraper(LegistarScraper):
    PYTHON_MUNICIPALITY_SLUG: str = "seattle"

    def __init__(self):
        """
        Seattle specific implementation of LegistarScraper.
        """
        super().__init__(
            client="seattle",
            timezone="America/Los_Angeles",
            ignore_minutes_item_patterns=[
                "This meeting also constitutes a meeting of the City Council",
                "In-person attendance is currently prohibited",
                "Times listed are estimated",
                "has been cancelled",
                "Deputy City Clerk",
                "Executive Sessions are closed to the public",
                # Sometimes will have number after "Session", e.g. "Session I"
                r"Executive Session \S*\s*on Pending, Potential, or Actual Litigation",
                "Items of Business",
                # Common to see "CITY COUNCIL:",
                # Or more generally "{body name}:"
                # Check for last char ":"
                r".+:$",
                "Pursuant to Washington State",
            ],
            static_data=parse_static_file(STATIC_FILE_DEFAULT_PATH),
            person_aliases=PERSON_ALIASES,
        )

        try:
            urlopen("https://seattlechannel.org/")
        except URLError:
            pass
        else:
            raise Exception(
                "seattlechannel.org may have fixed their SSL cert. "
                "Check and fix 'requests.get(*, verify=False)' calls"
            )

    def parse_content_uris(
        self, video_page_url: str, event_short_date: str
    ) -> List[ContentURIs]:
        """
        Return URLs for videos and captions parsed from seattlechannel.org web page

        Parameters
        ----------
        video_page_url: str
            URL to a web page for a particular meeting video

        event_short_date: str
            datetime representing the meeting's date, used for verification m/d/yy

        Returns
        -------
        content_uris: List[ContentURIs]
            List of ContentURIs objects for each session found.

        See Also
        --------
        get_content_uris()
        """
        with warnings.catch_warnings():
            warnings.simplefilter(
                "ignore",
                category=urllib3.exceptions.InsecureRequestWarning,
            )
            # now load the page to get the actual video url
            soup = BeautifulSoup(
                requests.get(video_page_url, verify=False).text,
                "html.parser",
            )

        # <script>
        # ...
        # playerInstance.setup({
        # sources: [
        #     {
        #         file: "//...mp4",
        #         label: "Auto"
        #     }
        # ],
        # ...
        # tracks: [{
        #     file: "documents/seattlechannel/closedcaption/2021/...vtt",
        #     label: "English",
        #     kind: "captions",
        #     "default": true
        # }
        #
        #  ],
        # ...

        # entire script tag text that has the video player setup call
        video_script_block = soup.find(
            "script", text=re.compile(r"playerInstance\.setup")
        )
        if not video_script_block:
            log.warning(
                f"Couldn't find 'playerInstance.setup()' block on {video_page_url}.\n"
                "seattlechannel.org may have changed their video page html"
            )
            return []
        video_script_text = video_script_block.string

        # halt if event date not in video's idstring
        # likely means some change on video web page source / script

        # e.g. idstring:'Select Budget Committee Session II 10/14/21'
        #      idstring:'City Council 10/11/21'
        if not re.search(f"idstring:.+{event_short_date}.+", video_script_text):
            raise ValueError(
                f"event date {event_short_date} not in video idstring.\n"
                f"{video_page_url} may be for a different event's video.\n"
            )

        # playerSetup({...
        #             ^
        player_arg_start = re.search(
            r"playerInstance\.setup\((\{)", video_script_text
        ).start(1)

        # ...});
        #     ^
        # playerInstance... # more playerInstance code
        video_json_blob = video_script_text[
            player_arg_start : player_arg_start
            + re.search(
                r"\)\;\s*\n\s*playerInstance", video_script_text[player_arg_start:]
            ).start(0)
        ]

        # not smart enough to make one-line regex for all the 'file's in 'sources'
        videos_start = video_json_blob.find("sources:")
        videos_end = video_json_blob.find("],", videos_start)
        # as shown above, url will start with // so prepend https:
        video_uris = [
            "https:" + i
            for i in re.findall(
                r"file\:\s*\"([^\"]+)",
                video_json_blob[videos_start:videos_end],
            )
        ]

        captions_start = video_json_blob.find("tracks:")
        captions_end = video_json_blob.find("],", captions_start)
        caption_uris = [
            "https://www.seattlechannel.org/" + i
            for i in re.findall(
                r"file\:\s*\"([^\"]+)",
                video_json_blob[captions_start:captions_end],
            )
        ]

        # use max count between videos and captions
        # so we don't lose any (e.g. caption = None if < # videos)
        iter = range(max(len(video_uris), len(caption_uris)))
        list_uri = []

        for i in iter:
            # just in case # videos != # captions
            try:
                video_uri = video_uris[i]
            except IndexError:
                video_uri = None

            try:
                caption_uri = caption_uris[i]
            except IndexError:
                caption_uri = None

            list_uri.append(ContentURIs(video_uri=video_uri, caption_uri=caption_uri))

        if len(list_uri) == 0:
            log.debug(f"No video URI found on {video_page_url}")
        return list_uri

    @staticmethod
    def roman_to_int(roman: str):
        """
        Roman numeral to an integer

        Parameters
        ----------
        roman: str
            Roman numeral string

        Returns
        -------
        int
            Input roman numeral as integer

        References
        ----------
        https://www.w3resource.com/python-exercises/class-exercises/python-class-exercise-2.php
        """
        rom_val = {"I": 1, "V": 5, "X": 10, "L": 50, "C": 100, "D": 500, "M": 1000}
        int_val = 0
        for i in range(len(roman)):
            if i > 0 and rom_val[roman[i]] > rom_val[roman[i - 1]]:
                # subtract twice the i-1 th number since it has already been added
                int_val += rom_val[roman[i]] - 2 * rom_val[roman[i - 1]]
            else:
                int_val += rom_val[roman[i]]
        return int_val

    def get_video_page_urls(
        self, video_list_page_url: str, event_short_date: str
    ) -> List[str]:
        """
        Return URLs to web pages hosting videos for meetings from event_short_date

        Parameters
        ----------
        video_list_page_url: str
            URL to web page listing videos featuring the responsible group/body
            for the event described in legistar_ev.
            e.g. http://www.seattlechannel.org/BudgetCommittee?Mode2=Video

        event_short_date: str
            datetime representing the meeting's date m/d/yy

        Returns
        -------
        video_page_urls: List[str]
            web page URL per video

        See Also
        --------
        get_content_uris()
        """
        with warnings.catch_warnings():
            warnings.simplefilter(
                "ignore",
                category=urllib3.exceptions.InsecureRequestWarning,
            )
            # request list of videos for this group on this event's date
            response = requests.get(
                # this is the query sent by the "filter" button on the web page
                f"{video_list_page_url}&filterTerm={quote_plus(event_short_date)}"
                "&itemsPerPage=25&toggleDisplay=Thumbnail_Excerpt",
                verify=False,
            ).text

        # <div class="paginationContainer">
        #     <div class="row borderBottomNone paginationItem">
        #         <div class="col-xs-12 col-sm-4 col-md-3">
        #                     <a href='/BudgetCommittee?videoid=x132213'... </a>
        #         </div>
        #         <div class="col-xs-12 col-sm-8 col-md-9">
        #             <div class="titleDateContainer">
        #                 <h2 class="paginationTitle">
        #                     <a href="/BudgetCommittee?videoid=x132213" ... </a>
        #             </h2>
        #                     <div class="videoDate">10/14/2021</div>
        #         </div>
        #         <div class="titleExcerptText"><p><em>Pursuant to Washington ... </div>
        #     </div>
        # </div>
        #    <div class="row borderBottomNone paginationItem">

        session_video_page_urls: Dict[int, str] = {}

        # want <a> tag in the <div> with
        # title attribute that contains the event date,
        # onclick attribute that calls loadJWPlayer,
        # href attribute that contains videoid

        soup = BeautifulSoup(response, "html.parser")
        for link in soup.find("div", class_="paginationContainer",).find_all(
            "a",
            href=re.compile("videoid"),
            onclick=re.compile("loadJWPlayer"),
            title=re.compile(event_short_date),
        ):
            # e.g. "Session I m/d/yy"
            match = re.search(
                r"session\s(?P<session_int>\d*)(?P<session_roman>[IVXLCDM]*)",
                link["title"],
                re.IGNORECASE,
            )
            if match:
                if match.group("session_int"):
                    session_video_page_urls[
                        int(match.group("session_int"))
                    ] = f"https://www.seattlechannel.org{link['href']}"
                elif match.group("session_roman"):
                    session_video_page_urls[
                        int(SeattleScraper.roman_to_int(match.group("session_roman")))
                    ] = f"https://www.seattlechannel.org{link['href']}"
                else:
                    session_video_page_urls[
                        len(session_video_page_urls)
                    ] = f"https://www.seattlechannel.org{link['href']}"

        # ordered by session number
        return [
            session_video_page_urls[session]
            for session in sorted(session_video_page_urls.keys())
        ]

    def get_content_uris(self, legistar_ev: Dict) -> List[ContentURIs]:
        """
        Return URLs for videos and captions parsed from seattlechannel.org web page

        Parameters
        ----------
        legistar_ev: Dict
            Data for one Legistar Event.

        Returns
        -------
        content_uris: List[ContentURIs]
            List of ContentURIs objects for each session found.

        See Also
        --------
        parse_content_uris()

        Notes
        -----
        get_events() calls get_content_uris() to get video and caption URIs.
        get_content_uris() gets video page URL from EventInSiteURL.
        If "videoid" in video page URL, calls parse_content_uris().
        Else, calls get_video_page_urls() to get proper video page URL with "videoid",
            then calls parse_content_uris().
        get_events()
            -> get_content_uris()
                -> parse_content_uris()
                or
                -> get_video_page_urls(), parse_content_uris()
        """
        try:
            # a td tag with a certain id pattern containing url to video
            with urlopen(legistar_ev[LEGISTAR_EV_SITE_URL]) as resp:
                soup = BeautifulSoup(resp.read(), "html.parser")
        except URLError or HTTPError:
            log.debug(f"Failed to open {legistar_ev[LEGISTAR_EV_SITE_URL]}")
            return []

        try:
            # this gets us the url for the web PAGE containing the video
            video_page_url = soup.find(
                "a",
                id=re.compile(r"ct\S*_ContentPlaceHolder\S*_hypVideo"),
                class_="videolink",
            )["href"]
            log.debug(f"{legistar_ev[LEGISTAR_EV_SITE_URL]} -> {video_page_url}")
        # catch if find() didn't find video web page url (no <a id=... href=.../>)
        except KeyError:
            log.debug(f"No URL for video page on {legistar_ev[LEGISTAR_EV_SITE_URL]}")
            return []

        event_short_date = datetime.fromisoformat(legistar_ev[LEGISTAR_SESSION_DATE])
        # want no leading zero for month or day
        event_short_date = (
            f"{event_short_date.month}/"
            f"{event_short_date.day}/"
            f"{event_short_date.strftime('%y')}"
        )

        # Some meetings will have text like "Session II" in "Meeting location".
        # For those, don't bother verifying video page URL.
        # They are multi-session and we need to call get_video_page_urls()
        if (
            "session ii"
            not in soup.find(
                "span", id=re.compile(r"ctl\S*_ContentPlaceHolder\S*_lblLocation$")
            ).text.lower()
        ):
            try:
                if parse_qs(urlsplit(video_page_url).query)["videoid"]:
                    # video link contains specific videoid
                    return self.parse_content_uris(video_page_url, event_short_date)
            except KeyError:
                pass

        # at this point video_page_url points to generic video list page like
        # http://www.seattlechannel.org/BudgetCommittee?Mode2=Video

        return [
            uris
            # 1 web page per session video for this multi-session event
            for page_url in self.get_video_page_urls(video_page_url, event_short_date)
            # video and caption urls on the session video web page
            for uris in self.parse_content_uris(page_url, event_short_date)
        ]

    @staticmethod
    def get_person_picture_url(person_www: str) -> Optional[str]:
        """
        Parse person_www and return banner image used on the web page

        Parameters
        ----------
        person_www: str
            e.g. http://www.seattle.gov/council/pedersen

        Returns
        -------
        Image URL: Optional[str]
            Full URL to banner image displayed on person_www
        """
        try:
            with urlopen(person_www) as resp:
                soup = BeautifulSoup(resp.read(), "html.parser")
        except URLError or HTTPError:
            log.debug("Failed to open {person_www}")
            return None

        # <div class="featureWrapperShort" style="background-image:
        # url('/assets/images/Council/Members/Pedersen/
        # Councilmember-Alex-Pedersen_homepage-banner.jpg')"></div>
        div = soup.find(
            "div", class_="featureWrapperShort", style=re.compile(r"background\-image")
        )
        if not div:
            return None

        try:
            # now get just the image uri '/assets/...'
            return "http://www.seattle.gov/" + re.search(
                r"url\('([^']+)", div["style"]
            ).group(1)
        except AttributeError:
            pass

        return None

    @staticmethod
    def get_static_person_info() -> Optional[List[Person]]:
        """
        Return partial Persons with static long-term information

        Returns
        -------
        persons: Optional[List[Person]]
        """
        try:
            # has table with all council members
            with urlopen("https://seattle.legistar.com/MainBody.aspx") as resp:
                soup = BeautifulSoup(resp.read(), "html.parser")
        except URLError or HTTPError:
            log.debug("Failed to open https://seattle.legistar.com/MainBody.aspx")
            return None

        static_person_info: List[Person] = []

        # <tr id="ctl00_ContentPlaceHolder1_gridPeople_ctl00__0" ...>
        #     <td class="rgSorted" style="white-space:nowrap;">
        #         <a ...>Alex Pedersen</a>
        #     </td>
        #     <td>Councilmember<br /><em>Council Position No. 4</em></td>
        #     <td>1/1/2020</td>
        #     <td style="white-space:nowrap;">
        #         <span ...>12/31/2023</span>
        #     </td>
        #     <td style="white-space:nowrap;">
        #         <a ...>Alex.Pedersen@seattle.gov</a>
        #     </td>
        #     <td style="white-space:nowrap;">
        #         <a ...>http://www.seat...ouncil/pedersen</a>
        #     </td>
        # </tr>
        for tr in soup.find_all(
            "tr",
            # each row with this id in said table is for a council member
            id=re.compile(r"ctl\d+_ContentPlaceHolder\d+_gridPeople_ctl\d+__\d+"),
        ):
            # <a> tag in this row with this id has full name
            try:
                name = str_simplified(
                    tr.find(
                        "a",
                        id=re.compile(
                            r"ctl\d*_ContentPlaceHolder\d*"
                            r"_gridPeople_ctl\d*_ctl\d*_hypPerson"
                        ),
                    ).text
                )
            except AttributeError:
                # find() returned None
                continue

            # <a> tag in this row with this id has url
            # for web page with more info on this person
            try:
                person_picture_url = SeattleScraper.get_person_picture_url(
                    tr.find(
                        "a",
                        id=re.compile(
                            r"ctl\d*_ContentPlaceHolder\d*"
                            r"_gridPeople_ctl\d*_ctl\d*_hypWebSite"
                        ),
                    )["href"]
                )
            except AttributeError:
                # find() returned None
                continue

            # <td> in this row with <br> and <em> has seat name
            # <td>Councilmember<br /><em>Council Position No. 4</em></td>
            # the seat is the <em>-phasized text
            try:
                seat = Seat(
                    name=str_simplified(
                        [
                            td
                            for td in tr.find_all("td")
                            if td.find("br") is not None and td.find("em") is not None
                        ][0].em.text
                    )
                )
            except IndexError:
                # accessed 0-th item in an empty list []
                continue

            # from "Council Position No. 4"
            #     Seat.electoral_area: District 4
            #     Seat.name: Position 4
            # from "At-large Council Position No. 9"
            #     Seat.electoral_area: At-large
            #     Seat.name: Position 9

            match = re.search(
                r"(?P<atlarge>At.*large)?.*position.*(?P<position_num>\d+)",
                seat.name,
                re.IGNORECASE,
            )
            if match:
                seat_number = match.group("position_num")
                seat.electoral_area = f"District {seat_number}"
                if match.group("atlarge"):
                    seat.electoral_area = "Citywide"

                seat.name = f"Position {seat_number}"

            static_person_info.append(
                Person(name=name, picture_uri=person_picture_url, seat=seat)
            )

        return static_person_info

    @staticmethod
    def dump_static_info(file_path: str) -> bool:
        """
        Save static data in json format

        Parameters
        ----------
        file_path: str
            Static data dump file path

        Returns
        -------
        bool
            True if some data was saved in file_path

        See Also
        --------
        LegistarScraper.inject_known_data()
        """
        static_person_info = {}
        for person in SeattleScraper.get_static_person_info():
            # save this Person in json keyed by the name
            static_person_info[person.name] = json.loads(person.to_json())

        if not static_person_info:
            return False

        with open(file_path, "wt") as dump:
            dump.write(
                json.dumps({STATIC_FILE_KEY_PERSONS: static_person_info}, indent=4)
            )
        return True



import json
import re
from copy import deepcopy
from datetime import datetime, timedelta
from itertools import filterfalse, groupby
from logging import getLogger
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Set

import cleantext
import pytz
from cdp_backend.database.constants import RoleTitle
from cdp_backend.pipeline.ingestion_models import (
    Body,
    IngestionModel,
    Person,
    Role,
    Seat,
)
from cdp_backend.utils.constants_utils import get_all_class_attr_values

from .types import ScraperStaticData

###############################################################################

log = getLogger(__name__)

###############################################################################


def reduced_list(input_list: List[Any], collapse: bool = True) -> Optional[List]:
    """
    Remove all None items from input_list.

    Parameters
    ----------
    input_list: List[Any]
        Input list from which to filter out items that are None
    collapse: bool, default = True
        If True, return None in place of an empty list

    Returns
    -------
    reduced_list: Optional[List]
        All items in the original list except for None values.
        None if all items were None and collapse is True.
    """
    filtered = [item for item in input_list if item is not None]
    if collapse and len(filtered) == 0:
        filtered = None

    return filtered


def str_simplified(input_str: str) -> str:
    """
    Remove leading and trailing whitespaces, simplify multiple whitespaces, unify
    newline characters.

    Parameters
    ----------
    input_str: str

    Returns
    -------
    cleaned: str
        input_str stripped if it is a string
    """
    if not isinstance(input_str, str):
        return input_str

    input_str = cleantext.clean(
        input_str, fix_unicode=True, lower=False, to_ascii=False
    )

    # Replace utf-8 char encodings with single utf-8 chars themselves
    input_str = input_str.encode("utf-8").decode("utf-8")

    return input_str


def parse_static_person(
    person_json: Dict[str, Any],
    all_seats: Dict[str, Seat],
    primary_bodies: Dict[str, Body],
) -> Person:
    """
    Parse Dict[str, Any] for a person in static data file to a Person instance.
    person_json["seat"] and person_json["roles"] are validated against
    all_seats and primary_bodies in static data file.

    Parameters
    ----------
    person_json: Dict[str, Any]
        A dictionary in static data file with info for a Person.

    all_seats: Dict[str, Seat]
        Seats defined as top-level in static data file

    primary_bodies: Dict[str, Body]
        Bodies defined as top-level in static data file.


    See Also
    --------
    parse_static_file()
    sanitize_roles()
    """
    log.debug(f"Begin parsing static data for {person_json['name']}")

    person: Person = Person.from_dict(
        # "seat" and "roles" are not direct serializations of Seat/Role
        {k: v for k, v in person_json.items() if k != "seat" and k != "roles"}
    )
    if "seat" not in person_json:
        log.debug("Seat name not given")
        return person

    seat_name: str = person_json["seat"]
    if seat_name not in all_seats:
        log.error(f"{seat_name} is not defined in top-level 'seats'")
        return person

    # Keep all_seats unmodified; we will append Roles to this person.seat below
    person.seat = deepcopy(all_seats[seat_name])
    if "roles" not in person_json:
        log.debug("Roles not given")
        return person

    # Role.title must be a RoleTitle constant so get all allowed values
    role_titles: List[str] = get_all_class_attr_values(RoleTitle)
    for role_json in person_json["roles"]:
        if (
            # if str, it is looked-up in primary_bodies
            isinstance(role_json["body"], str)
            and role_json["body"] not in primary_bodies
        ):
            log.error(
                f"{role_json} is ignored. "
                f"{role_json['body']} is not defined in top-level 'primary_bodies'"
            )
        elif role_json["title"] not in role_titles:
            log.error(
                f"{role_json} is ignored. "
                f"{role_json['title']} is not a RoleTitle constant."
            )
        else:
            role: Role = Role.from_dict(
                {k: v for k, v in role_json.items() if k != "body"}
            )
            if isinstance(role_json["body"], str):
                role.body = primary_bodies[role_json["body"]]
            else:
                # This role.body is a dictionary and defines a non-primary one
                # e.g. like a committee such as Transportation
                # that is not the main/full council
                role.body = Body.from_dict(role_json["body"])

            if person.seat.roles is None:
                person.seat.roles = [role]
            else:
                person.seat.roles.append(role)

    return person


def parse_static_file(file_path: Path) -> ScraperStaticData:
    """
    Parse Seats, Bodies and Persons from static data JSON

    Parameters
    ----------
    file_path: Path
        Path to file containing static data in JSON

    Returns
    -------
    ScraperStaticData:
        Tuple[Dict[str, Seat], Dict[str, Body], Dict[str, Person]]

    See Also
    -----
    parse_static_person()
    sanitize_roles()

    Notes
    -----
    Function looks for "seats", "primary_bodies", "persons" top-level keys
    """
    with open(file_path) as static_file:
        static_json: Dict[str, Dict[str, Any]] = json.load(static_file)

        if "seats" not in static_json:
            seats: Dict[str, Seat] = {}
        else:
            seats: Dict[str, Seat] = {
                seat_name: Seat.from_dict(seat)
                for seat_name, seat in static_json["seats"].items()
            }

        if "primary_bodies" not in static_json:
            primary_bodies: Dict[str, Body] = {}
        else:
            primary_bodies: Dict[str, Body] = {
                body_name: Body.from_dict(body)
                for body_name, body in static_json["primary_bodies"].items()
            }

        if "persons" not in static_json:
            known_persons: Dict[str, Person] = {}
        else:
            known_persons: Dict[str, Person] = {
                person_name: parse_static_person(person, seats, primary_bodies)
                for person_name, person in static_json["persons"].items()
            }

        log.debug(
            f"ScraperStaticData parsed from {file_path}:\n"
            f"    seats: {list(seats.keys())}\n"
            f"    primary_bodies: {list(primary_bodies.keys())}\n"
            f"    persons: {list(known_persons.keys())}\n"
        )
        return ScraperStaticData(
            seats=seats, primary_bodies=primary_bodies, persons=known_persons
        )


def sanitize_roles(
    person_name: str,
    roles: Optional[List[Role]] = None,
    static_data: Optional[ScraperStaticData] = None,
    council_pres_patterns: List[str] = ["chair", "pres", "super"],
    chair_patterns: List[str] = ["chair", "pres"],
) -> Optional[List[Role]]:
    """
    1. Standardize roles[i].title to RoleTitle constants
    2. Ensure only 1 councilmember Role per term

    Parameters
    ----------
    person_name: str
        Sanitization target Person.name

    roles: Optional[List[Role]] = None
        target Person's Roles to sanitize

    static_data: Optional[ScraperStaticData]
        Static data defining primary council bodies and predefined Person.seat.roles.
        See Notes.

    council_pres_patterns: List[str]
        Set roles[i].title as "Council President" if match
        and roles[i].body is a primary body like City Council
    chair_patterns: List[str]
        Set roles[i].title as "Chair" if match
        and roles[i].body is not a primary body

    Notes
    -----
    Remove roles[#] if roles[#].body in static_data.primary_bodies.
    Use static_data.persons[#].seat.roles instead.

    If roles[i].body not in static_data.primary_bodies,
    roles[i].title cannot be "Councilmember" or "Council President".

    Use "City Council" and "Council Briefing"
    if static_data.primary_bodies is empty.
    """
    if roles is None:
        roles = []

    if not static_data or not static_data.primary_bodies:
        # Primary/full council not defined in static data file.
        # these are reasonably good defaults for most municipalities.
        primary_body_names = ["city council", "council briefing"]
    else:
        primary_body_names = [
            body_name.lower() for body_name in static_data.primary_bodies.keys()
        ]

    try:
        have_primary_roles = len(static_data.persons[person_name].seat.roles) > 0
    except (KeyError, AttributeError, TypeError):
        have_primary_roles = False

    def _is_role_period_ok(role: Role) -> bool:
        """
        Test that role.[start | end]_datetime is acceptable
        """
        if role.start_datetime is None or role.end_datetime is None:
            return False
        if not have_primary_roles:
            # no roles in static data; accept if this this role is current
            return role.start_datetime.astimezone(
                pytz.utc
            ) <= datetime.today().astimezone(pytz.utc) and datetime.today().astimezone(
                pytz.utc
            ) <= role.end_datetime.astimezone(
                pytz.utc
            )
        # accept if role coincides with one given in static data
        for static_role in static_data.persons[person_name].seat.roles:
            if (
                static_role.start_datetime <= role.start_datetime
                and role.end_datetime <= static_role.end_datetime
            ):
                return True
        return False

    def _is_primary_body(role: Role) -> bool:
        """
        Is role.body one of primary_bodies in static data file
        """
        return (
            role.body is not None
            and role.body.name is not None
            and str_simplified(role.body.name).lower() in primary_body_names
        )

    def _fix_primary_title(role: Role) -> str:
        """
        Council president or Councilmember
        """
        if (
            role.title is None
            or re.search(
                "|".join(council_pres_patterns), str_simplified(role.title), re.I
            )
            is None
        ):
            return RoleTitle.COUNCILMEMBER
        return RoleTitle.COUNCILPRESIDENT

    def _fix_nonprimary_title(role: Role) -> str:
        """
        Not council president or councilmember
        """
        if role.title is None:
            return RoleTitle.MEMBER

        role_title = str_simplified(role.title).lower()
        # Role is not for a primary/full council
        # Role.title cannot be Councilmember or Council President
        if "vice" in role_title:
            return RoleTitle.VICE_CHAIR
        if "alt" in role_title:
            return RoleTitle.ALTERNATE
        if "super" in role_title:
            return RoleTitle.SUPERVISOR
        if re.search("|".join(chair_patterns), role_title, re.I) is not None:
            return RoleTitle.CHAIR
        return RoleTitle.MEMBER

    def _is_councilmember_term(role: Role) -> bool:
        return (
            role.title == RoleTitle.COUNCILMEMBER
            and role.start_datetime is not None
            and role.end_datetime is not None
        )

    roles = list(
        # drop dynamically scraped primary roles
        # if primary roles are given in static data
        filterfalse(
            lambda role: have_primary_roles and _is_primary_body(role),
            # filter out bad start_datetime, end_datetime
            filter(_is_role_period_ok, roles),
        )
    )
    # standardize titles
    for role in filter(_is_primary_body, roles):
        role.title = _fix_primary_title(role)
    for role in filterfalse(_is_primary_body, roles):
        role.title = _fix_nonprimary_title(role)

    class CouncilMemberTerm(NamedTuple):
        start_datetime: datetime
        end_datetime: datetime
        index_in_roles: int

    # when checking for overlapping terms, we should do so per body.
    # e.g. simultaneous councilmember roles in city council and in council briefing
    # are completely acceptable and common.

    scraped_member_roles_by_body: List[List[Role]] = [
        list(roles_for_body)
        for body_name, roles_for_body in groupby(
            sorted(
                filter(
                    # get all dynamically scraped councilmember terms
                    lambda role: not have_primary_roles
                    and _is_councilmember_term(role),
                    roles,
                ),
                # sort from old to new role
                key=lambda role: (
                    role.body.name,
                    role.start_datetime,
                    role.end_datetime,
                ),
            ),
            # group by body
            key=lambda role: role.body.name,
        )
    ]

    if have_primary_roles:
        # don't forget to include info from the static data file
        roles.extend(static_data.persons[person_name].seat.roles)
    if len(scraped_member_roles_by_body) == 0:
        # no Councilmember roles dynamically scraped
        # nothing more to do
        return roles

    for roles_for_body in scraped_member_roles_by_body:
        for i in [i for i, role in enumerate(roles_for_body) if i > 0]:
            prev_role = roles_for_body[i - 1]
            this_role = roles_for_body[i]
            # if member role i overlaps with member role j, end i before j
            if prev_role.end_datetime > this_role.start_datetime:
                roles[
                    roles.index(prev_role)
                ].end_datetime = this_role.start_datetime - timedelta(days=1)

    return roles


class IngestionModelScraper:
    """
    Base class for events scrapers providing IngestionModels for cdp-backend pipeline

    Parameters
    ----------
    timezone: str
        The timezone for the target client.
        i.e. "America/Los_Angeles" or "America/New_York"
        See https://en.wikipedia.org/wiki/List_of_tz_database_time_zones for canonical
        timezones.
    person_aliases: Optional[Dict[str, Set[str]]]
        Dictionary used to catch name aliases
        and resolve improperly different Persons to the one correct Person.
        Default: None
    """

    def __init__(
        self,
        timezone: str,
        person_aliases: Optional[Dict[str, Set[str]]] = None,
    ):
        self.timezone: pytz.timezone = pytz.timezone(timezone)
        self.person_aliases = person_aliases

    @staticmethod
    def find_time_zone() -> str:
        """
        Return name for a US time zone matching UTC offset calculated from OS clock.
        """
        utc_now = pytz.utc.localize(datetime.utcnow())
        local_now = datetime.now()

        for zone_name in pytz.country_timezones("us"):
            zone = pytz.timezone(zone_name)
            # if this is my time zone
            # utc_now as local time should be VERY close to local_now
            if (
                abs(
                    (
                        utc_now.astimezone(zone) - zone.localize(local_now)
                    ).total_seconds()
                )
                < 5
            ):
                return zone_name

        return None

    def localize_datetime(self, local_time: datetime) -> datetime:
        """
        Return input datetime with time zone information.
        This allows for nonambiguous conversions to other zones including UTC.

        Parameters
        ----------
        local_time: datetime
            The datetime to attached timezone information to.

        Returns
        -------
        local_time: datetime
            The date and time attributes (year, month, day, hour, ...) remain unchanged.
            tzinfo is now provided.
        """
        try:
            return self.timezone.localize(local_time)
        except (AttributeError, ValueError):
            # AttributeError: time_zone or local_time is None
            # ValueError: local_time is not navie (has time zone info)
            return local_time

    @staticmethod
    def get_required_attrs(model: IngestionModel) -> List[str]:
        """
        Return list of keys required in model as specified in IngestionModel class
        definition.

        Parameters
        ----------
        model: IngestionModel
            Person, MinutesItem, etc.

        Returns
        -------
        attr_keys: List[str]
            List of keys (attributes) in model without default value in class
            definition.
        """
        try:
            # create an empty one to have python tell us what keys are required
            model.__class__()
            # all attrs in model have default values
            return []
        except TypeError as e:
            # e.g. __init__() missing 3 required positional arguments:
            # 'session_datetime', 'video_uri', and 'session_index'
            match = re.search(
                r"missing (?P<num_keys>\d+) required.+argument(?:s)?\:\s*(?P<keys>.+)",
                str(e),
            )

        if not match:
            log.debug(f"not able to get required attributes for {model.__class__}")
            return []

        num_keys = int(match.group("num_keys"))

        # 'session_datetime', 'video_uri', and 'session_index'
        # -> ["session_datetime", "video_uri", "session_index"]

        # SHOULD be able to do this more elegantly using re.split()
        # but couldn't quite get the pattern right
        keys = re.sub(
            # TypeError uses
            # , and
            # and
            # ,
            # as delimiters for attribute names
            r"(\s*,\s*and\s*)|(\s*and\s*)|(\s*,\s*)",
            ",",
            match.group("keys").strip().replace("'", ""),
        ).split(",")

        if num_keys != len(keys):
            log.debug(f"{model.__class__} has {num_keys} required keys but got {keys}")

        return keys

    def get_none_if_empty(self, model: IngestionModel) -> Optional[IngestionModel]:
        """
        Check required keys in model, return None if any such key has no value.
        i.e. If all required keys have valid value, return as-is.

        Parameters
        ----------
        model: IngestionModel
            Person, MinutesItem, etc.

        Returns
        -------
        model: Optional[IngestionModel]
            None or model as-is
        """
        try:
            min_keys = self.min_ingestion_keys[model.__class__]
        except AttributeError:
            # first time using min_ingestion_keys
            self.min_ingestion_keys = {}
            min_keys = None
        except KeyError:
            # first time checking model.__class__
            min_keys = None

        if min_keys is None:
            min_keys = IngestionModelScraper.get_required_attrs(model)
            # cache so we don't do expensive dynamic checking
            # again for this IngestionModel
            self.min_ingestion_keys[model.__class__] = min_keys

        if not min_keys:
            # no required keys for this model
            # this probably never happens
            return model

        for key in min_keys:
            try:
                val = getattr(model, key)

                # "if not" test to catch all None and None-like values
                # e.g. empty string, empty list, ...
                # but int(0) is not "empty"
                if not val and not isinstance(val, int):
                    # empty value for this key in model
                    return None
            except AttributeError:
                return None

        # nonempty value for all required keys in model
        return model

    def resolve_person_alias(self, person: Person) -> Person:
        """
        If input person is in fact an alias of a reference known person,
        return the reference person instead.
        Else return person as-is.

        Parameters
        ----------
        person: Person
            Person to check whether is an alias or a real unique Person

        Returns
        -------
        Person
            input person, or the correct reference Person if input person is an alias.
            This base implementation always returns person as-is.

        See Also
        --------
        instances.seattle.person_aliases
        """
        return person



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from bs4 import BeautifulSoup
from typing import Optional, List
from urllib.parse import unquote
from urllib.request import urlopen
from defusedxml import ElementTree
import re

from .types import ContentURIs, LegistarContentParser
from .scraper_utils import str_simplified


def _parse_format_1(client: str, soup: BeautifulSoup) -> Optional[List[ContentURIs]]:
    """
    Works for milwaukee, kingcounty, phoenix

    Parameters
    ----------
    client: str
        Which legistar client to target. Ex: "seattle"
    video web page: BeautifulSoup
        Video web page loaded into bs4

    Returns
    -------
    uris: Optional[List[ContentURIs]]
        URIs for video and optional caption

    See Also
    --------
    cdp_scrapers.types.LegistarContentParser
    cdp_scrapers.legistar_utils.get_legistar_content_uris()
    """
    # source link for the video is embedded in the script of downloadLinks.
    # <script type="text/javascript">
    # var meta_id = '',
    # currentClipIndex = 0,
    # clipList = eval([8844]),
    # downloadLinks = eval([["\/\/69.5.90.100:443\/MediaVault\/Download.aspx?
    # server=king.granicus.com&clip_id=8844",
    # "http:\/\/archive-media.granicus.com:443\/OnDemand\/king\/king_e560cf63-5570-416e-a47d-0e1e13652224.mp4",null]]);
    # </script>

    video_script_text = soup.find("script", text=re.compile(r"downloadLinks"))
    if video_script_text is None:
        return None

    video_script_text = video_script_text.string
    # Below two lines of code tries to extract video url from downLoadLinks variable
    # "http:\/\/archive-media.granicus.com:443\/OnDemand\/king\/king_e560cf63-5570-416e-a47d-0e1e13652224.mp4"
    try:
        downloadLinks = video_script_text.split("[[")[1]
        video_url = downloadLinks.split('",')[1].strip('"')
    except IndexError:
        # split() did not yield expected # items
        return None
    # Cleans up the video url to remove backward slash(\)
    video_uri = video_url.replace("\\", "")
    # caption URIs are not found for kingcounty events.
    return [ContentURIs(video_uri=video_uri, caption_uri=None)]


def _parse_format_2(client: str, soup: BeautifulSoup) -> Optional[List[ContentURIs]]:
    """
    Works for denver

    Parameters
    ----------
    client: str
        Which legistar client to target. Ex: "seattle"
    video web page: BeautifulSoup
        Video web page loaded into bs4

    Returns
    -------
    uris: Optional[List[ContentURIs]]
        URIs for video and optional caption

    See Also
    --------
    cdp_scrapers.types.LegistarContentParser
    cdp_scrapers.legistar_utils.get_legistar_content_uris()
    """
    # <div id="download-options">
    # <a href="...mp4">
    video_url = soup.find("div", id="download-options")
    if video_url is None:
        return None
    return [ContentURIs(str_simplified(video_url.a["href"]))]


def _parse_format_3(client: str, soup: BeautifulSoup) -> Optional[List[ContentURIs]]:
    """
    Works for boston, corpuschristi, elpasotexas

    Parameters
    ----------
    client: str
        Which legistar client to target. Ex: "seattle"
    video web page: BeautifulSoup
        Video web page loaded into bs4

    Returns
    -------
    uris: Optional[List[ContentURIs]]
        URIs for video and optional caption

    See Also
    --------
    cdp_scrapers.types.LegistarContentParser
    cdp_scrapers.legistar_utils.get_legistar_content_uris()
    """
    # <video>
    # <source src="...">
    # <track src="...">
    video_url = soup.find("video")
    if video_url is None:
        return None
    return [
        ContentURIs(
            video_uri=f"https:{str_simplified(video_url.source['src'])}",
            caption_uri=(
                (
                    f"http://{client}.granicus.com/"
                    f"{str_simplified(video_url.track['src'])}"
                )
                # transcript is nice to have but not required
                if video_url.find("track") is not None
                and "src" in video_url.track.attrs
                else None
            ),
        )
    ]


def _parse_format_4(client: str, soup: BeautifulSoup) -> Optional[List[ContentURIs]]:
    """
    Works for longbeach, richmondva

    Parameters
    ----------
    client: str
        Which legistar client to target. Ex: "seattle"
    video web page: BeautifulSoup
        Video web page loaded into bs4

    Returns
    -------
    uris: Optional[List[ContentURIs]]
        URIs for video and optional caption

    See Also
    --------
    cdp_scrapers.types.LegistarContentParser
    cdp_scrapers.legistar_utils.get_legistar_content_uris()
    """
    # a long <meta content="...VideoUrl=...&..." />
    url_regex = re.compile("VideoUrl=([^&]+)")
    # TODO: Also constains ScriptURL but appears to be invalid.
    # Double-check against more events; scrape if we can.
    player_meta = soup.find("meta", property="og:video", content=url_regex)
    if not player_meta:
        return None
    video_url = f"https:{unquote(url_regex.search(player_meta['content']).group(1))}"
    # this makes the server return an xml-like asx
    # NOTE: changing query from rtmp to http makes it return the video url
    #       in http as opposed to rtmp, e.g. rtmp://...mp4 -> http://...mp4
    video_url = video_url.replace("stream_type=rtmp", "stream_type=http")

    with urlopen(video_url) as resp:
        asx = ElementTree.fromstring(resp.read())
        # one and only one <REF HREF="http://...mp4" />
        ref_tag = asx.find(".//REF")
        return [ContentURIs(ref_tag.get("HREF"))] if ref_tag is not None else None


# TODO: do dynamically using inspect or something similar
all_parsers: List[LegistarContentParser] = [
    _parse_format_1,
    _parse_format_2,
    _parse_format_3,
    _parse_format_4,
]
